close all
clear 
clc

% ======= (1) TRANSMITTED SIGNALS ======= %
% Signal source directions
az = [-10;-35;55]; % azimuth
el = zeros(size(az))+90; %elevation (not considered in this case)
R = [5;7;10];
c = physconst('LightSpeed');
fs = 77e9;
lambda = c/fs;
Ts = 50e-6; %sweep time (single chirp is transmitted)
BW = 4e9;
cr = BW/Ts; 
td = 2*R/c;
fb = cr*td;
M = length(az); % Number of reflectors
L = 8000; % ADCsamples, Number of data snapshots recorded by the receiver 
t = linspace(0,Ts,L);
m = zeros(M,L);

for i = 1:M
    m(i,:) = cos(2*pi*t*fb(i,1));
end
% ========= (2) RECEIVED SIGNAL ========= %
% Wavenumber vectors (in units of wavelength/2)
k = pi*[sind(az).*sind(el), cosd(az).*sind(el), cosd(el)].';
% Array geometry [rx,ry,rz]
N = 40; % Number of virtual antennas
r = [(-(N-1)/2:(N-1)/2).',zeros(N,2)]; % Assume uniform linear array
% Matrix of array response vectors
A = exp(-1j*r*k);
% Additive noise
sigma2 = 0.01; % Noise variance
n = sqrt(sigma2)*(randn(N,L) + 1j*randn(N,L))/sqrt(2);
% Received signal
x = A*m + n;

for i = 1:N
    fft_x (i,:) = 1/L*fftshift((fft(x(i,:),L)));
end
Fs = 1/(t(3)-t(2)); % Sampling frequency, t(3)-t(2) is the time step
df=Fs/L; %frequency resolution
sampleIndex = -L/2:L/2-1; % ordered index for FFT plot
f=sampleIndex*df; % x-axis index converted to ordered frequencies
rng = f*Ts*c/(2*BW); % x-axis index converted to range

figure
hold on
for i=1:10
    plot(rng(L/2+1:end), abs(fft_x(i,L/2+1:end))) % positive frequencies
end
axis([0 15 0 0.5])
xlabel('Range (m)')
ylabel('Amplitude')
% ========= (3) MUSIC ALGORITHM ========= %
% Sample covariance matrix
Rxx = x*x'/L;
d = 0.5;% Distance between elements in Wavelenght
[Q ,D]=eig(Rxx); %Compute eigendecomposition of covariance matrix
[D,I]=sort(diag(D),1,'descend'); %Find r largest eigenvalues
Q=Q (:,I); %Sort the eigenvectors to put signal eigenvectors first
Qs=Q (:,1:M); %Get the signal eigenvectors
Qn=Q(:,M+1:N); %Get the noise eigenvectors
% MUSIC algorithm
% Define angles at which MUSIC “spectrum” will be computed
angles=(-90:1:90);
for k=1:length(angles)
a1(:,k)=exp(-1i*2*pi*(d*(0:N-1)'*sin([angles(k).']*pi/180)));
end
for k=1:length(angles)
%Compute MUSIC “spectrum”
music_spectrum(k)=(a1(:,k)'*a1(:,k))/(a1(:,k)'*(Qn*Qn')*a1(:,k));
end
% figure
% plot(angles,abs(music_spectrum)/max(abs(music_spectrum)))
% grid on
% title('MUSIC Spectrum')
% xlabel('Angle in degrees')
% ========= (4) Bartlett Beamformer ========= %
for k=1:length(angles)
%Compute Bartlett "spectrum"
Bartlett_spectrum(k)=(a1(:,k)'*Rxx*a1(:,k));
end
% figure
% hold on
% plot(angles,abs(Bartlett_spectrum)/max(abs(Bartlett_spectrum)))
% grid on
% title('Bartlett Spectrum')
% xlabel('Angle in degrees')

% ========= (5) Capon Beamformer ========= %
for k=1:length(angles)
%b(:,k)=(inv(Rxx)*a1(:,k))/(a1(:,k)'*inv(Rxx)*a1(:,k)); 
b(:,k)=(Rxx\a1(:,k))/(a1(:,k)'*(Rxx\a1(:,k))); %Same as previous line
end
for k=1:length(angles)
%Compute Capon "spectrum"
Capon_spectrum(k)=(b(:,k)'*Rxx*b(:,k));
end
% figure
% hold on
% plot(angles,abs(Capon_spectrum)/max(abs(Capon_spectrum)))
% grid on
% title('Capon Spectrum')
% xlabel('Angle in degrees')

% ========= (6) ESPRIT ========= %
Qs0 = Qs(1:N-1,:);
Qs1 = Qs(2:N,:);
[Psi,rnk]=linsolve(Qs1,Qs0);
[C ,Phi]=eig(Psi);
doas = rad2deg(asin(angle(diag(Phi)) /(2*pi*d)))

% ========= (7) Maximum Likelihood (Deterministic) ========= %
%P = A * ((A' * A) \ A');
for k=1:length(angles)
    P = a1(:,k) * (((a1(:,k))' * a1(:,k)) \ (a1(:,k))');
    DML(k,1) = trace(P * Rxx);
    %DML(k,1) = trace((eye(N)-P) * Rxx);
end
% figure
% plot(angles,abs(DML)/max(abs(DML)))
% grid on
% title('Deterministic Maximum Likelihood Spectrum')
% xlabel('Angle in degrees')

% ========= (8) Maximum Likelihood (Stochastic) ========= %
for k=1:length(angles)
    P = a1(:,k) * (((a1(:,k))' * a1(:,k)) \ (a1(:,k))');
    SML(k,1) =  log10(det((P * Rxx * P') +(1/(N-M)) * (trace(eye(N) - P)) * (eye(N) - P*P')));
end
% figure
% plot(angles,abs(SML)/max(abs(SML)))
% grid on
% title('Stochastic Maximum Likelihood Spectrum')
% xlabel('Angle in degrees')

% ========= Plot and Compare Methods ========= %
figure
hold on
plot(angles,abs(Bartlett_spectrum)/max(abs(Bartlett_spectrum)))
plot(angles,abs(Capon_spectrum)/max(abs(Capon_spectrum)))
plot(angles,abs(DML)/max(abs(DML)))
plot(angles,abs(SML)/max(abs(SML)))
plot(angles,abs(music_spectrum)/max(abs(music_spectrum)))
for i = 1:length(doas)
    plot(doas(i),1,'o')
end
box on
grid on
title('DOA Estimation Methods')
xlabel('Angle (Degrees)')
ylabel('Normalized Output')
legend('Bartlett','MVDR','DML','SML','MUSIC','ESPRIT')
axis([-90 90 0 1.01])
pbaspect([2 1 1])


%% Range-Azimuth heatmap

NUM_ANGLE_BINS = 1000;
x1 = fft(x,NUM_ANGLE_BINS);
theta = asind([-NUM_ANGLE_BINS/2+1 : NUM_ANGLE_BINS/2-1]'*(2/NUM_ANGLE_BINS));
%range = 0:c/(2*BW):f(end)*c/(2*cr);

x1 = x1';
x1 = fftshift(abs(x1));
x1=flip(x1,2);
x2 = fft(x1,length(x1));
x2 = x2';
QQ = fftshift(abs(x2),2);
QQ=QQ.';

figure
imagesc(theta,rng(L/2+2:end),abs(QQ(L/2+2:end,:)))
title('Gives the wrong range and azimuth')
axis([-80 80 8 22])
xlabel('Angle')
ylabel('Range')

figure
imagesc(theta/1.43,rng(L/2+2:end)/2,abs(QQ(L/2+2:end,:)))
title('Gives the right range and azimuth')
axis([-60 60 4 12])
xlabel('Angle')
ylabel('Range')

figure
hold on
for i=1:NUM_ANGLE_BINS
    plot(theta,abs(x1(i,1:end-1)))
end
xlabel('Angle')
ylabel('Amplitude')
title ('One dimensional angle-FFT along the 40 virtual antennas, Correct estimation')

